/* -------------------------------------------------------
                         charlie20.h

     Header fuer Softwaremodul zur Ansteuerung von
     20 LED's mittels Charlieplexing

     MCU   :  STM8S103F3
     Takt  :  interner oder extern

     20.03.2018  R. Seelig
   ------------------------------------------------------ */

/*
##################################################################
                      STM8S103F3P6 Pinout
##################################################################


                            ------------
UART1_CK / TIM2_CH1 / PD4  |  1     20  |  PD3 / AIN4 / TIM2_CH2 / ADC_ETR
    UART1_TX / AIN5 / PD5  |  2     19  |  PD2 / AIN3
    UART1_RX / AIN6 / PD6  |  3     18  |  PD1 / SWIM
                     NRST  |  4     17  |  PC7 / SPI_MISO
              OSCIN / PA1  |  5     16  |  PC6 / SPI_MOSI
             OSCOUT / PA2  |  6     15  |  PC5 / SPI_CLK
                Vss (GND)  |  7     14  |  PC4 / TIM1_CH4 / CLK_CCO / AIN2
                VCAP (*1)  |  8     13  |  PC3 / TIM1_CH3 /
                Vdd (+Ub)  |  9     12  |  PB4 / I2C_SCL
           TIM2_CH3 / PA3  | 10     11  |  PB5 / I2C_SDA
                            -----------

*1 :  Ist mit min. 1uF gegen GND zu verschalten

*/

/* -------------------------------------------------------------------------------
                        Charlieplexing 20-LED Modul

     Beim Charlieplexing sind jeweils 2 antiparallel geschaltete Leuchtdioden
     an 2 GPIO-Leitungen angeschlossen. Hieraus ergeben sich 10 Paare zu jeweils
     2 LEDs (bei 5 verwendeten GPIO-Anschluessen).

     Bsp.:

      A  o------+-----,         C o------+-----,
                |     |                  |     |
               ---   ---                ---   ---
             A \ /   / \ B            C \ /   / \ D
             B ---   --- A            D ---   --- C
                |     |                  |     |
      B  o------+-----'         D o------+-----'


      A  B  B  C  C  D  D  E  A  C  C  E  D  B  A  D  A  E  E  B   Linenkombination
      B  A  C  B  D  C  E  D  C  A  E  C  B  D  D  A  E  A  B  E
      ----------------------------------------------------------
      O  O  O  O  O  O  O  O  O  O  O  O  O  O  O  O  O  O  O  O   LED
      ----------------------------------------------------------
      0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19  LED-Nummern
   ------------------------------------------------------------------------------- */

#ifndef in_charlie20
  #define in_charlie20

  #include <stdint.h>
  #include "stm8_init.h"
  #include "stm8_gpio.h"
  #include "stm8s.h"

  #define  pindefs     1                // 0: Konfiguration Steckbrett
                                        // 1: Konfiguration mit RDA5807 UKW-Receiver

/* -------------------------------------------------------
     wichtigte globale Variable:
     Ein Beschreiben dieser Variablen fuehrt bei durch-
     gefuehrte Initialisierung zum Anzeigen des Bitmusters
     des Inhaltes von charlie20_buf
   ------------------------------------------------------- */
  extern uint32_t charlie20_buf;                  // Buffer in dem ein Bitmuster aufgenommen wird,
                                                  // welches ueber den Timerinterrupt ausgegeben wird

  extern const uint16_t cplex[20];

/* -------------------------------------------------------
                        Prototypen
   ------------------------------------------------------- */
  void tim2_ovf(void) __interrupt(13);
  void tim2_init(void);
  void charlie20_allinput(void);
  void charlie20_init(void);
  void charlie20_lineset(char nr);

  // Zuordnung LEDs zu GPIO-Anschluessen, Makros zum Setzen / Ruecksetzen von einzelnen
  // GPIO-Leitungen

  // Steckbrettkonfiguration
  #if (pindefs == 0)

    #define charlieA_output()  PD4_output_init()
    #define charlieA_input()   PD4_input_init()
    #define charlieA_set()     PD4_set()
    #define charlieA_clr()     PD4_clr()

    #define charlieB_output()  PD3_output_init()
    #define charlieB_input()   PD3_input_init()
    #define charlieB_set()     PD3_set()
    #define charlieB_clr()     PD3_clr()

    #define charlieC_output()  PD2_output_init()
    #define charlieC_input()   PD2_input_init()
    #define charlieC_set()     PD2_set()
    #define charlieC_clr()     PD2_clr()

    #define charlieD_output()  PD1_output_init()
    #define charlieD_input()   PD1_input_init()
    #define charlieD_set()     PD1_set()
    #define charlieD_clr()     PD1_clr()

    #define charlieE_output()  PC3_output_init()
    #define charlieE_input()   PC3_input_init()
    #define charlieE_set()     PC3_set()
    #define charlieE_clr()     PC3_clr()

  #endif

  // LED-Anzeige fuer RDA5807 UKW-Receiver
  #if (pindefs == 1)

    #define charlieA_output()  PC7_output_init()
    #define charlieA_input()   PC7_input_init()
    #define charlieA_set()     PC7_set()
    #define charlieA_clr()     PC7_clr()

    #define charlieB_output()  PC6_output_init()
    #define charlieB_input()   PC6_input_init()
    #define charlieB_set()     PC6_set()
    #define charlieB_clr()     PC6_clr()

    #define charlieC_output()  PC5_output_init()
    #define charlieC_input()   PC5_input_init()
    #define charlieC_set()     PC5_set()
    #define charlieC_clr()     PC5_clr()

    #define charlieD_output()  PC4_output_init()
    #define charlieD_input()   PC4_input_init()
    #define charlieD_set()     PC4_set()
    #define charlieD_clr()     PC4_clr()

    #define charlieE_output()  PC3_output_init()
    #define charlieE_input()   PC3_input_init()
    #define charlieE_set()     PC3_set()
    #define charlieE_clr()     PC3_clr()

  #endif

  #define charlie20_bufset(nr)   (charlie20_buf |=  ( (uint32_t)1 << (uint32_t)nr ))
  #define charlie20_bufclr(nr)   (charlie20_buf &= ~( (uint32_t)1 << (uint32_t)nr ))

#endif
