/* -------------------------------------------------------
                         charlie20.c

     Softwaremodul zur Ansteuerung von 20 LED's mittels
     Charlieplexing

     MCU   :  STM8S103F3
     Takt  :  interner oder extern

     20.03.2018  R. Seelig
   ------------------------------------------------------ */

/* -------------------------------------------------------------------------------
                        Charlieplexing 20-LED Modul

     Beim Charlieplexing sind jeweils 2 antiparallel geschaltete Leuchtdioden
     an 2 GPIO-Leitungen angeschlossen. Hieraus ergeben sich 10 Paare zu jeweils
     2 LEDs.

     Bsp.:

      A  o------+-----,         C o------+-----,
                |     |                  |     |
               ---   ---                ---   ---
             A \ /   / \ B            C \ /   / \ D
             B ---   --- A            D ---   --- C
                |     |                  |     |
      B  o------+-----'         D o------+-----'


      A  B  B  C  C  D  D  E  A  C  C  E  D  B  A  D  A  E  E  B   Linenkombination
      B  A  C  B  D  C  E  D  C  A  E  C  B  D  D  A  E  A  B  E
      ----------------------------------------------------------
      ----------------------------------------------------------
      0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16 17 18 19  LED-Nummern
   ------------------------------------------------------------------------------- */

/*
##################################################################
                      STM8S103F3P6 Pinout
##################################################################


                            ------------
UART1_CK / TIM2_CH1 / PD4  |  1     20  |  PD3 / AIN4 / TIM2_CH2 / ADC_ETR
    UART1_TX / AIN5 / PD5  |  2     19  |  PD2 / AIN3
    UART1_RX / AIN6 / PD6  |  3     18  |  PD1 / SWIM
                     NRST  |  4     17  |  PC7 / SPI_MISO
              OSCIN / PA1  |  5     16  |  PC6 / SPI_MOSI
             OSCOUT / PA2  |  6     15  |  PC5 / SPI_CLK
                Vss (GND)  |  7     14  |  PC4 / TIM1_CH4 / CLK_CCO / AIN2
                VCAP (*1)  |  8     13  |  PC3 / TIM1_CH3 /
                Vdd (+Ub)  |  9     12  |  PB4 / I2C_SCL
           TIM2_CH3 / PA3  | 10     11  |  PB5 / I2C_SDA
                            -----------

*1 :  Ist mit min. 1uF gegen GND zu verschalten

##################################################################
   STM8S103F3P6 Minimal-Board (China)  alternative Pinfunktionen
##################################################################

Hinweis: VCAP ist NICHT auf dem Board aufgelegt !!!

                                            +---+
                                            |   |
                                       +----|   |----+
                                       |    +---+    |
  PD4 / UART1_CK / TIM2_CH1 / BEEP(HS) | o         o | D3 / HS / AIN4 / TIM2_CH2 / ADC_ETR
            PD5 / UART1_TX / AIN5 / HS | o         o | PD2 / HS / AIN3 / TIM2_CH3
             PD6 / UART1_RX / AIN6 / H | o         o | PD1 / HS / SWIM ( ST-Link )
             NRST ( auch fuer ST-Link) | o         o | PC7 / HS / SPI_MISO / TIM1_CH2
                   PA1 / OSCIN (Quarz) | o         o | PC6 / HS / SPI_MOSI / TIM1_CH1
                  PA2 / OSCOUT (Ouarz) | o         o | PC5 / HS / SPI_SCK
                                   GND | o         o | PC4 / HS / TIM1_CH4 / CLK_CC0 / AIN2 / TIM1_CH2N
                                   +5V | o         o | PC3 / HS / TIM1_CH3 / TLI / TIM1_CH1N
                                 +3,3V | o         o | PB4 / T / I2C_SCL / ADC_ETR
   PA3 / SPI_NSS / TIM2_CH3(HS)        | o         o | PB5 / T / I2C_SDA / TIM1_BKIN
                                       |   o o o o   |
                                       +-------------+

                                       nrst gnd swim Vcc
*/


#include "../include/charlie20.h"


uint32_t charlie20_buf= 0;             // Buffer in dem ein Bitmuster aufgenommen wird,
                                       // welches ueber den Timerinterrupt ausgegeben wird


const uint16_t cplex[20] =
  // HiByte: definiert, welche der Charlieplexingleitungen Ausgaenge sind
  // LoByte: definiert, welche Bits gesetzt oder geloescht sind
  //
  //    D15 D14 D13 D12 D11 D10 D09 D08 || D07 D06 D05 D04 D03 D02 D01 D00
  //    ------------------------------------------------------------------
  //                  E   D   C   B   A                  E   D   C   B   A
  //
  // Bsp.: 0x0302

  //   A und B sind Ausgaenge, B= 0, A= 1 => LED AB leuchtet

  {
  //  0=AB    1=BA    2=BC    3=CB    4=CD    5=DC    6=DE    7=ED
    0x0301, 0x0302, 0x0602, 0x0604, 0x0c04, 0x0c08, 0x1808, 0x1810,
  //  8=AC    9=CA   10=CE    11=EC  12=DB   13=BD   14=AD   15=DA
    0x0501, 0x0504, 0x1404, 0x1410, 0x0a08, 0x0a02, 0x0901, 0x0908,
  // 16=AE   17=EA   18=EB    19=BE
    0x1101, 0x1110, 0x1210, 0x1202
  };

/* ------------------------------------------------------
                         tim2_ovf

     ISR fuer Timer2 overflow.

     Hier werden 20 Bits des Buffers charlie20_buf im
     Zeitmultiplex nacheinander dargestellt.
   ------------------------------------------------------ */
void tim2_ovf(void) __interrupt(13)
// 13 ist der Interruptvektor fuer Timer3 overflow
{
  static uint8_t isr_cnt = 0;

  charlie20_lineset(20);                         // alle LED's aus
  if (charlie20_buf & (1ul << isr_cnt))          // Bit in charlie20_buf gesetzt ?
  {
    charlie20_lineset(isr_cnt);                  // dann diese LED einschalten
  }
  isr_cnt++;
  isr_cnt= isr_cnt % 20;

  TIM2_SR1 &= ~TIM2_SR1_UIF;                     // Interrupt quittieren
}


/* ------------------------------------------------------
                     tim2_init
      Initialisiert Timer2 als Overflow-Interrupt-Timer
      (Interruptvector 13)

      Berechnung:

      tim2_takt  = F_CPU / 2 ^ scalvalue
      time       = tim2_takt * (relvalue -1)

      Bsp.:
      scalvalue = 0;         Teilertakt / 1 => 0.0625 uS
      relvalue  = 15999;     time= (relvalue+1) * scalvalue = 16000 * 0.0625 uS
                                                            = 1 mS



      retvalue  = 7999
      scalvalue = 0
      =>          0.5 mS

      retvalue  = 15625
      scalvalue = 9
      =>          0.5 s

      retvalue = 319
      scalvalue = 9
      =>          10 ms

   ------------------------------------------------------ */
void tim2_init(void)
{

  #define relvalue    7999
  #define scalvalue   0

  TIM2_PSCR= scalvalue;                  // F_CPU / 2 ^  scalvalue
  TIM2_IER |= TIM2_IER_UIE;              // enable overflow
  TIM2_CR1 |= TIM2_CR1_CEN;              // Timer2 enable ( CEN = ClockENable )
  TIM2_CR1 |= TIM2_CR1_ARPE;             // Timer2 autoreload ( mit Werten in TIM2_ARR )

  TIM2_ARRH= (uint8_t) (relvalue >> 8);
  TIM2_ARRL= (uint8_t) (relvalue & 0x00ff);

  int_enable();
}


/* ------------------------------------------------------
                   charlie20_allinput

     schaltet alle am Charlieplexing beteiligten I/O
     Leitungen als Eingang
   ------------------------------------------------------ */
void charlie20_allinput(void)
{
  charlieA_input();
  charlieB_input();
  charlieC_input();
  charlieD_input();
  charlieE_input();
}

/* ------------------------------------------------------
                      charlie20_init

     initialisiert die I/O Leitungen des Charlieplexings
     und setzt alle Leitungen auf 1
   ------------------------------------------------------ */
void charlie20_init(void)
{
  charlie20_allinput();
  tim2_init();
}

/* ------------------------------------------------------
                    charlie20_lineset

     schaltet eine einzelne LED an.

     nr:   Nummer der einzuschaltenden LED (0..19)
           nr== 20 => alle LED's werden ausgeschaltet
   ------------------------------------------------------ */
void charlie20_lineset(char nr)
{
  uint8_t   bl, bh;

  charlie20_allinput();
  if (nr== 20) return;                               // 20 = alle LED aus

  bh= cplex[nr] >> 8;
  bl= cplex[nr] &0x00ff;

  if (bh & 0x01)
  {
    charlieA_output();
    if (bl & 0x01) charlieA_set(); else charlieA_clr();
  }
  if (bh & 0x02)
  {
    charlieB_output();
    if (bl & 0x02) charlieB_set(); else charlieB_clr();
  }
  if (bh & 0x04)
  {
    charlieC_output();
    if (bl & 0x04) charlieC_set(); else charlieC_clr();
  }
  if (bh & 0x08)
  {
    charlieD_output();
    if (bl & 0x08) charlieD_set(); else charlieD_clr();
  }
  if (bh & 0x10)
  {
    charlieE_output();
    if (bl & 0x10) charlieE_set(); else charlieE_clr();
  }
}

